/*******************

FileOps.c : Performs the basic File Operations

Get Attributes.
Create a File/Directory
Open a File
Read a File
Write to a File
File Seek
Close a  File
Delete reference to the File
Read a Directory
Delete reference to the directory

*********************/
#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<string.h>
#include "LibDisk.h"
#include "metadata.h"





/**lfs_getattr: get file/directory attributes by reading corresponding inode

   args: path of the file,
   pointer to the location that will hold the values
   returns : -1 on failure
   0 on success
**/
int lfs_getattr(char *path,inode_hello *inode){
  int inodeno;
  inodeno=traverse_path(path);
  if(inodeno==-1)
    return -1;
  if(read_inode(inodeno,inode,Rd)!=0)
    return -1;
    return 0;
}

/**lfs_filecreate: creates a file (maps to mknod)
   
   args: path of the file,
   permissions with which the file will be created

   returns: -1 on failure
   0 on success
   
**/
int lfs_filecreate(char *path,perms permissions,inode_type itype){
  
  int inodeno,steps=0,icount=0,flag=0,block,parentblock,incr,value=-1;
  int *inodenoptr;
  inode_hello *inodenew,*inodeparent;
  char path1[256],buf[SECTOR_SIZE];
  char **temp=(char **)malloc(sizeof(char *)*300);
  inodenew=(inode_hello *)malloc(sizeof(inode_hello));
  inodeparent=(inode_hello *)malloc(sizeof(inode_hello));
  strcpy((char *)path1,(char *)path);
  if((strlen(path1)==0)||(strlen(path1)>255))
    return -1;
  temp[steps]=strtok(path1,"/");
  while(temp[steps]!=NULL){
    steps++;
    temp[steps]=strtok(NULL,"/");
  }
  
  inodeno=traverse_path(path);
  parentblock=parent;
  if(inodeno!=-1){
    if(bitmap_update(I,inodeno,1,Rd)==1)
      return -1;
  }

  inodenew->name=(char *)malloc(sizeof(char)*16);
  strncpy(inodenew->name,temp[steps-1],strlen(temp[steps-1]));
  inodenew->type=itype;
  inodenew->size=0;
  for(icount=0;icount<30;icount++)
    inodenew->blockptr[icount]=-1;
  inodenew->mode=permissions;

  inodeno=find_free(I);
  
  if(read_inode(inodeno,inodenew,Up)!=0)
    return -1;
  
  if(read_inode(parent,inodeparent,Rd)!=0)
    return -1;
  if(inodeparent->type!=D)
    return -1;
  for(icount=0;icount<30;icount++){
    if(inodeparent->blockptr[icount]==-1){
      if(icount==0){
	block=find_free(B);
	inodeparent->blockptr[icount]=block;
	inodeparent->size+=1;
	if(bitmap_update(B,block,1,Up)!=0)
	  return -1;
	break;
      }else{
	block=inodeparent->blockptr[icount-1];
	if(Disk_Read(parentblock,buf)!=0)
	  return -1;
	incr=0;
	while(incr<481){
	  int value=(int)(*(buf+incr));
	  if(value==0){
	    flag=1;
	    break;
	  }
	  incr++;			
	}
	if(flag!=1)
	  block=find_free(B);
	inodeparent->blockptr[icount]=block;
	inodeparent->size+=1;
	if(bitmap_update(B,block,1,Up)!=0)
	  return -1;
	break;
	
      }
    }	
  }
  
  inodenoptr=&inodeno;
  if(flag==1){
    strncpy(buf+incr,(void *)temp[steps-1],sizeof(char)*16);
    strncpy(buf+incr+16,(void *)((char *)inodenoptr),sizeof(char));
    if(Disk_Write(block,buf)!=0)
      return -1;
  }else{
    if(Disk_Read(block,buf)!=0)
      return -1;
    strncpy(buf,(void *)temp[steps-1],sizeof(char)*16);
    strncpy(buf+16,(void *)((char *)inodenoptr),sizeof(char));
    if(Disk_Write(block,buf)!=0)
      return -1;
  }

  if(read_inode(parentblock,inodeparent,Up)!=0)
    return -1;	  
  if(bitmap_update(I,inodeno,1,Up)!=0)
    return -1;

  free(temp);
  free(inodenew);
  free(inodeparent);										
  return inodeno;
}

/**
   lfs_openfile: Open File

   args: Path Name
   
   return: file descriptor on success
   -1 on failure
   
**/
int lfs_openfile(char *path){
  int inodeno,count=0,flag=-1,fd;
  inode_hello *check;
  check = (inode_hello *)malloc(sizeof(inode_hello));
  inodeno=traverse_path(path);
  if(inodeno==-1)
    return -1;
  if(read_inode(inodeno,check,Rd)!=0)
    return -1;
  if(check->type!=F)
    return -1;   
  while(count<25){
    if(fopentable[count][1]==0){
      flag=1;
      fd=count;
      fopentable[count][1]=inodeno;
      fopentable[count][2]=0;
      break;
    }
  }
  if(flag==-1)
    return -1;
  free(check);
  return fd;
}

	
/**
   lfs_readfile: Read from a file
   
   args: file descriptor
   buffer
   size to be read
      
   returns: bytes read on success
   -1 on failure
**/
int lfs_readfile(int fd,void *buffer,int size){
  int inodeno,count=0,flag=-1,filepointer,newloc,sizer,newblock,newbyte;
  int offnumofblocks,remainderblocks,sizeblocks,endbytes,block,blockfile;
  char buf[SECTOR_SIZE];
  inode_hello *inoderead;
  if(fopentable[fd][-1]==-1)
    return -1; // file not open

  if(size>SECTOR_SIZE)
    return -1;
  
  inoderead=(inode_hello *)malloc(sizeof(inode_hello));
  if(read_inode(inodeno,inoderead,Rd)!=0)
    return -1; //error while reading the inode
  
  sizeblocks=(inoderead->size)/SECTOR_SIZE;
  endbytes=(inoderead->size)%SECTOR_SIZE;
  
  offnumofblocks=fopentable[fd][2]/512;
  remainderblocks=fopentable[fd][2]%512;
  
  newloc=fopentable[fd][2]+size;
  if(newloc>inoderead->size){
    size=inoderead->size-fopentable[fd][2];
    newloc=inoderead->size;
  }
  
  newblock=newloc/SECTOR_SIZE;
  newbyte=newloc%SECTOR_SIZE;
  block=inoderead->blockptr[offnumofblocks];
  if(Disk_Read(block,buf)!=0)
    return -1;
  if(size>(SECTOR_SIZE-remainderblocks)){
    sizer=SECTOR_SIZE-remainderblocks;
    memcpy((void *)buffer,(void *)(buf+remainderblocks),sizeof(char)*sizer);
    memset(buf,0,sizeof(char)*SECTOR_SIZE);
    block=inoderead->blockptr[newblock];
    if(Disk_Read(block,buf)!=0)
      return -1;
    memcpy((void *)(buffer+sizer),(void *)buf,sizeof(char)*(size-sizer));
}else{
    memcpy((void *)buffer,(void *)(buf+remainderblocks),sizeof(char)*size);
  }
  
  
fopentable[fd][2]=newloc;
 
 free(inoderead);
 return size;
}
  

/** 
    lfs_writefile : Writes to a file
    
    args: file descriptor
    buffer
    sixe to be written
    
    returns :size written on success
    -1 on failure
    
**/

int lfs_writefile(int fd,void *buffer,int size){
  int inodeno,count,newloc,block,offnumofblocks,remainderblocks,sizer,ret;
  inode_hello *inodewrite;
  char buf[512]={0};
  inodeno=fopentable[fd][1];
  if(inodeno==-1)
    return -1;
  if(size>SECTOR_SIZE)
    return -1;
  inodewrite=(inode_hello *)malloc(sizeof(inode_hello));
  if(read_inode(inodeno,inodewrite,Rd)!=0)
    return -1;
  newloc=fopentable[fd][2]+size;
  if(newloc>MAXFILESIZE)
    return -1;

  offnumofblocks=fopentable[fd][2]/SECTOR_SIZE;
  remainderblocks=fopentable[fd][2]%SECTOR_SIZE;
  
  
  block=inodewrite->blockptr[offnumofblocks];
  if(block==-1){
    block=find_free(B);
    inodewrite->blockptr[offnumofblocks]=block;
    if(bitmap_update(B,block,1,Up)!=0)
      return -1;
  }else{
    if(Disk_Read(block,buf)!=0)
      return -1;
  }
  if(size>(SECTOR_SIZE-remainderblocks)){
    sizer=SECTOR_SIZE-remainderblocks;
    memcpy((void *)(buf+remainderblocks),(void *)buffer,sizeof(char)*sizer);
    if(Disk_Write(block,buf)!=0)
      return -1;
    block=inodewrite->blockptr[offnumofblocks+1];
    if(block==-1){
      block=find_free(B);
      inodewrite->blockptr[offnumofblocks+1]=block;
      inodewrite->size+=size;
      if(bitmap_update(B,block,1,Up)!=0)
	return -1;
      if(read_inode(inodeno,inodewrite,Up)!=0){
	ret=bitmap_update(B,block,0,Up);
	return -1;
      }
    }
    
    if(Disk_Read(block,buf)!=0)
      return -1;
    memcpy((void *)buf,(void *)(buffer+sizer),sizeof(char)*(size-sizer));
    if(Disk_Write(block,buf)!=0)
      return -1;
  }else{
    memcpy((void *)(buf+remainderblocks),(void *)buffer,sizeof(char)*size);
    if(Disk_Write(block,buf)!=0)
      return -1;
    inodewrite->size+=size;
    if(read_inode(inodeno,inodewrite,Up)!=0){
	ret=bitmap_update(B,block,0,Up);
	return -1;
     }
  }  
  fopentable[fd][2]=newloc;
  free(inodewrite);
  return size;
}

/**
   lfs_fileseek: updates the file pointer

   args : file descriptor
   int offset
   
   returns -1 on failure
   new location of file pointer
**/
int lfs_fileseek(int fd,int offset){
  int inodeno;
  inode_hello *inodeseek;
  inodeseek=(inode_hello *)malloc(sizeof(inode_hello));
  inodeno=fopentable[fd][1];
  if(inodeno==-1)
    return -1;
  if(read_inode(inodeno,inodeseek,Rd)!=0)
    return -1;
  if(inodeseek->size<offset)
    return -1;
  if(offset>MAXFILESIZE)
    return -1;
  
  fopentable[fd][2]=offset;
  
  free(inodeseek);
  return offset;
  
}

/**
   
   lfs_fileclose: closes a open file
   
   args: file descriptor

   returns : -1 on failure
   0 on success
**/

int lfs_close(int fd){
  int inodeno;
  inodeno=fopentable[fd][1];
  if(inodeno==-1)
    return -1;
  fopentable[fd][1]=-1;
  fopentable[fd][2]=0;
  
  return 0;
}

/**
   lfs_fileunlink: deletes reference to the file

   args : pathname of file
   
   returns : -1 on failure
   return 0;
**/

int lfs_fileunlink(char *path){
  int inodeno,count=0,parentdir,flag=-1,block;
  inode_hello *inodeunlink,*parentd;
  char buf[512];
  char *filename;
    inodeno=traverse_path(path);
  parentdir=parent;
  if(inodeno==-1)
    return -1;
  for(count=0;count<25;count++){
    if(fopentable[count][1]==inodeno){
      break;
      return -1;
	}
  }
  parentd=(inode_hello *)malloc(sizeof(inode_hello));
  inodeunlink=(inode_hello *)malloc(sizeof(inode_hello));
  if(read_inode(parentdir,parentd,Rd)!=0)
    return -1;
  if(read_inode(inodeno,inodeunlink,Rd)!=0)
    return -1;
  
  filename=inodeunlink->name;
  
  for(count=0;count<30;count++){
    while(flag!=1){
      block=parentd->blockptr[count];
      if(Disk_Read(block,buf)!=0)
	return -1;
      for(count=0;count<26;count++){
	if(strncmp((void *)filename,(void *)(buf+(count*sizeof(direntry))),sizeof(char)*16)==0){
	memset((void *)(buf+(count*sizeof(direntry))),0,sizeof(direntry));
	flag=1;
	if(Disk_Write(block,buf)!=0)
	  return -1;
	parentd->size-=1;
	break;
	}
      }
    }
  }

  memset(buf,0,sizeof(char)*SECTOR_SIZE);
  for(count=0;count<30;count++){
    block=inodeunlink->blockptr[count];
    if(block!=-1){
      if(bitmap_update(B,block,0,Up)!=0)
	return -1;
      if(Disk_Write(block,buf)!=0)
	return -1;
      inodeunlink->blockptr[count]=-1;
    }
  }
  if(read_inode(parentdir,parentd,Up)!=0)
    return -1;
  
  if(bitmap_update(I,inodeno,0,Up)!=0)
    return -1;
  
  free(parentd);
  free(inodeunlink);
 
  return 0;
}


/**
   lfs_dirread: reads the contents of the directory
   
   args: path name
      buffer
      number of entries
      
      return -1 on failure
      number of bytes read on success

**/

int lfs_dirread(char *path,void *buffer,int size){
  int inodeno,count=0,inr=0,sizedir,block,incr;
  inode_hello *inodedir;
  char buf[512];
  
  inodeno=traverse_path(path);
  if(inodeno==-1)
    return -1;
  inodedir=(inode_hello *)malloc(sizeof(inode_hello));
  if(read_inode(inodeno,inodedir,Rd)!=0)
    return -1;
  if(inodedir->type!=D)
    return -1;
  sizedir=inodedir->size;
  if(sizedir>size)
    return -1;
  int numblock=(inodedir->size*sizeof(direntry))/500;
  int remainnumblock=(sizedir*sizeof(direntry))%500;
  for(count=0;count<numblock;count++){
    block=inodedir->blockptr[count];
    if(block==-1)
      break;
    if(Disk_Read(block,buf)!=0)
       return -1;
    for(incr=0;incr<26;incr++){
      
      memcpy((void *)(buffer+incr*(sizeof(direntry))+((count+1)*SECTOR_SIZE)), \
	     (void *)(buf+incr*(sizeof(direntry))),sizeof(direntry));
    }
  }

  block=inodedir->blockptr[numblock];
  if(Disk_Read(block,buf)!=0)
    return -1;
  memcpy((void *)(buffer+(numblock*500)),(void *)buf, sizeof(char)*remainnumblock);

  free(inodedir);
  return size;
}

/**
   lfs_dirunlink : Removing the reference to directory

   args: pathname 
   
   returns -1 on failure
   0 on success
**/
int lfs_dirunlink(char *path){
  int inodeno,parentdir,count=0,incr=0,flag=-1,block,parentsize;
  inode_hello *dir,*parentd;
  char buf[512],*dirname;
  inodeno=traverse_path(path);
  parentdir=parent;
  if(inodeno==-1)
    return -1;
  dir=(inode_hello *)malloc(sizeof(inode_hello));
  if(read_inode(inodeno,dir,Rd)!=0)
    return -1;
  if(dir->type!=D)
    return -1;
  if(dir->size>0)
    return -1;
  dirname=dir->name;
  parentd=(inode_hello *)malloc(sizeof(inode_hello));
  if(read_inode(parentdir,parentd,Rd)!=0)
    return -1;
  parentsize=parentd->size;
  for(count=1;count<=parentsize;count++){ 
    block=parentd->blockptr[count];
    if(Disk_Read(block,buf)!=0)
      return -1;
    for(incr=0;incr<26;incr++){
      if(strncmp(dirname,buf+(incr*sizeof(direntry)),sizeof(char)*16)==0){
	memset(buf+(incr*sizeof(direntry)),-1,sizeof(direntry));
	parentd->size-=1;
	flag =1;
	break;
      }
    }
  }
  
  if(read_inode(parentdir,parentd,Up)!=0)
    return -1;
  
  
  memset(buf,-1,sizeof(char)*SECTOR_SIZE);
  for(count=0;count<30;count++){
    block=dir->blockptr[count];
    if(block!=-1){
      if(Disk_Write(block,buf)!=0)
	return -1;
      dir->blockptr[count]=-1;
      if(bitmap_update(B,block,0,Up)!=0)
	return -1;
    }
  }
      
  if(read_inode(parentdir,parentd,Up)!=0)
    return -1;
  
  if(bitmap_update(I,inodeno,0,Up)!=0)
    return -1;

  free(dir);
  free(parentd);
  return 0;
}

  





  
      
  

  


  


  

    
    
  
  
      

  

        

	
	      	
     




	
	
		

	
	












