/** adminops.c-- Functions required to perform operations on Disk Image!

bitmap_update= used to update the inode and blk bitmap.
**/

#include "LibDisk.h"
#include "metadata.h"
#include<string.h>
#include<unistd.h>
#include<stdlib.h>
#include<stdio.h>

/**
bitmap_update : args: blocktype: Inode, Data
		      blkno    : Block number.
		      mask     : bit to be set as 1 or 0

return 0 on success,
       -1 on failure.   	
**/ 

int bitmap_update(blocktype blktype,int blkno,int mask,operation update){
  blocktype blktypefb=blktype;
  int blknum=blkno;
  char buf[512];
  int value=0;
  int *tempint;
  int byteloc,bitloc,sectorno=2,maskval;
  
  //intialize pointer
  tempint=&value;	
  
  //select right sector	
  if(blktype==I){
    sectorno=1;
  }else{
    blknum-=40;
    if(blknum<4096){
      sectorno=2;
    }else if(4095<blknum<8192){
      sectorno=3;
    }else{
      sectorno=4;
    }
  }
  //calculate the bit positions
  byteloc =(blknum/8);
  bitloc  =(blknum%8);
  // read sector from disk
  int ret_read=Disk_Read(sectorno,buf);
  
  memcpy((void *)tempint,(void *)(buf+byteloc),sizeof(char));
  value=(int)(*tempint);

 if(update==Up){
  	if(mask==1){
		maskval=1;
		maskval<<=bitloc;
		value|=maskval;
	  }else{
   		 maskval=254;
		 maskval<<=bitloc;
   		 value&=maskval;
		}
	  tempint=&value;
  	  memcpy((void *)(buf+byteloc),(void *)tempint,sizeof(char));
 	  if(Disk_Write(sectorno,buf)!=0)
		   return -1;
	  else
		   return 0;
  	}else{
		 value>>=bitloc;
		 return(value&1);
	}
		
 }

/**
read_inode : read/update inode!

**/

int read_inode(int inodeno, inode_hello *inode,operation update){
  int alloc_status,temp=0,sectorno=5,incr=0;
  char buf[512];

	if((bitmap_update(I,inodeno,temp,Rd))!=1 && update==Rd){
	      return -1;
	    }

  sectorno=INODES+(inodeno/3);
  //printf("\nsectorno checkpoint 1: %d, inodeno=%d" ,sectorno,inodeno);  
  if((inodeno%3)==0){
    incr=0;
  }else if((inodeno%3)==1){
    incr=1;
  }else{
    incr=2;
  }
  
  if(Disk_Read(sectorno,buf)!=0)
  	return -1;


  if(update==Rd){	
    memcpy((void *)inode,(void *)(buf+(incr*sizeof(inode_hello))),sizeof(inode_hello));
	return 0;
  }else{
	
    memcpy((void *)(buf+(incr*sizeof(inode_hello))),(void *)inode,sizeof(inode_hello));
	if(Disk_Write(sectorno,buf)!=0)
		return -1;
	return 0;

  }
    
  
}

/**

Find free : find free inode or data block 

**/

int find_free(blocktype blktype){
	int bitloc=0,byteloc=0,sectorno,value=0,maskval;
	char buf[SECTOR_SIZE];	
	//temp=&value;
	if(blktype==I){
		sectorno=1;
		if(Disk_Read(sectorno,buf)!=0)
			return -1;
		while(byteloc<13){
			
				value=(int)(*(buf+byteloc));
				for(bitloc=0;bitloc<8;bitloc++){
				if(byteloc!=12 && bitloc<4){
					maskval=value;
					maskval>>=bitloc;
			if(((maskval&1)==0) &&(((8*byteloc)+bitloc)!=101))
					return ((8*byteloc)+bitloc);
					}
				}
			byteloc++;
			}
			return -1;
		}else{
		sectorno=2;
		if(Disk_Read(sectorno,buf)!=0)
			return -1;
		while(byteloc<375){
				value=(int)(*(buf+byteloc));
                                for(bitloc=0;bitloc<8;bitloc++){
                                        maskval=value;
                                        maskval>>=bitloc;
                        	if((maskval&1)==0)
                                        return (((8*byteloc)+bitloc)+40);
                                 }
			byteloc++;
                      }
                   return -1;
	}

	return -1;
}
					
			
/**
traverse_path: traverses the path of file/directory and locates the inode 

args: path 
	
returns : -1 when error
         inodeno on success

**/
int traverse_path(char *path){
  char buf[SECTOR_SIZE];
  int icount=0,steps=0,inodeno=2,blkcount=0,flag=0,entr,length;
  char **temp=(char **)malloc(sizeof(char *)*300);
  length=strlen(path);
  char path1[256];
  inode_hello *inode=(inode_hello *)malloc(sizeof(inode_hello));

  strcpy((char *)path1,(char *)path);
  //path1[length+1]='\0';
  //memcpy((void *)path1,(void *)path,(sizeof(char)*strlen(path)));
  if((strlen(path1)==0)||(strlen(path1)>255))
    return -1;
  temp[steps]=strtok(path1,"/");
  while(temp[steps]!=NULL){
    steps++;
    temp[steps]=strtok(NULL,"/");
  }
  if(steps==0){
    inodeno=2;
    parent=2;
    return inodeno;
  }
  
  inodeno=2;
  for(icount=0;icount<steps;icount++){
    if(icount==steps-1){
      parent=inodeno;
    }		
    flag =0;
    if(bitmap_update(I,inodeno,0,Rd)==-1)
      return -1;
    
    if(read_inode(inodeno,inode,Rd)==-1)
      return -1;
    while(blkcount<30 || flag!=1){
      int blkptr=inode->blockptr[blkcount];
      if(blkptr==-1)
	break;
      if(Disk_Read(blkptr,buf)!=0)
	return -1;
      for(entr=0;entr<26;entr++){
	if(strncmp(temp[icount],buf+(sizeof(direntry)*entr),(sizeof(char)*16))==0){
	  inodeno=(int)(*(buf+(sizeof(char)*16)));
	  flag =1;
	  break;
	}
	
      }
	blkcount++;
    }
    if(flag !=1)
	  return -1;
  }
  return inodeno;
}

  
      
			
			
	




	
			
  

