/***

metadata.h : Defines limits and data structures for File system API

**/



#include<stdlib.h>
#include<stdio.h>
#include<unistd.h>

/* define the block size, the maximum number of files,
maximum block size and locations of various metadata structures */


#define MAXIMUM_FILE_NUMBER 100
#define MAXIMUM_FILE_SIZE 30*SECTOR_SIZE
#define SUPERBLOCK 0   
#define INODEBITMAP 1
#define BLOCKBITMAP 2
#define INODES 5
#define DATABLOCKS  40
#define ROOT_INODE 2
#define FILENAMELENGTH 16
#define MAXFILESIZE  SECTOR_SIZE*30

/* definition for inode type and permissions */
typedef enum {D,F}inode_type;
typedef enum {R,W,X,RW,RX,WX,RWX} perms;
typedef enum {I,B} blocktype;
typedef enum {Up,Rd} operation;

/* directory entry structure */
typedef struct directent{
	char name[16];
	int inodeno;
}direntry;

/* structure for superblock */

typedef struct superblk{
	char *magic;
	int freeinodes;
	int freeblocks;
	int min;
	int maxfiles;
	int maxopen;
	int blocksize;
	int reservedblk;
}superblock;

/* Inode structure */
typedef struct inode{
	char *name;
	inode_type type;
	int size;
	int blockptr[30];
	perms mode;
}inode_hello;

/*typedef struct opentab{
	inode_hello * inode;
	perms chkperms;
}open_table;*/

extern int osErrno;
int parent;
int fopentable[25][2];
    
// error types - don't change anything about these!! (even the order!)
typedef enum {
    E_GENERAL,      // general
    E_CREATE, 
    E_NO_SUCH_FILE, 
    E_TOO_MANY_OPEN_FILES, 
    E_BAD_FD, 
    E_NO_SPACE, 
    E_FILE_TOO_BIG, 
    E_SEEK_OUT_OF_BOUNDS, 
    E_FILE_IN_USE, 
    E_BUFFER_TOO_SMALL, 
    E_DIR_NOT_EMPTY,
    E_ROOT_DIR,
} FS_Error_t;
    
// File system generic call
int FS_Boot(char *path);
int bitmap_update(blocktype blktype,int blkno,int mask,operation update);
int read_inode(int inodeno,inode_hello *inode,operation update);
int find_free(blocktype blktype);
int traverse_path(char *path);
int lfs_getattr(char *path,inode_hello *inode);
int lfs_filecreate(char *path,perms permissions,inode_type itype);
int lfs_openfile(char *path);
int lfs_readfile(int fd,void *buffer,int size);
int lfs_writefile(int fd,void *buffer,int size);
int lfs_fileseek(int fd,int offset);
int lfs_close(int fd);
int lfs_fileunlink(char *file);
int lfs_dirread(char *path,void *buffer,int size);
int lfs_dirunlink(char *path);
int FS_Sync(char *path);


/* End of metadata header*/
	
        


	









