/*

testfuse.c :  test routines Interface between lfs FS and fuse for redirection

  gcc -Wall `pkg-config fuse --cflags --libs` fusexmp.c -o fusexmp
*/

#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/time.h>
#include "metadata.h"
#include "LibDisk.h"


/*** lfs_getattr : Fuse interface for the function getattr

returns : -err on Failure
          0 on Success

***/

static int fl_getattr(const char *path, struct stat *stbuf)
{
	int res=0;
        inode_hello *inodeattr;
	inodeattr=(inode_hello *)malloc(sizeof(inode_hello));
	memset(stbuf,0,sizeof(struct stat));
	
	if(lfs_getattr(path,inodeattr)!=0)
	  res = -ENOENT;
	if(inodeattr->type==D)
	  stbuf->st_mode=0755;
	else{
	  stbuf->st_mode=0666;
	  stbuf->st_size=inodeattr->size;
	}
	stbuf->st_link=1;
        free(inodeattr);
	return res;
}

/***

fl_readdir : fuse interface for reading from a directory

Returns : -1 for Failure
           0 on Success.

***/

static int fl_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
		       off_t offset, struct fuse_file_info *fi)
{
  int res=0,entrno=0,icount=0,*inodeno,ino=2;
  inode_hello *inodeattr,*inodeentry;
  char name[16]={0};

  (void)offset;
  (void)fi;

  inodeattr=(inode_hello *)malloc(sizeof(inode_hello));
  inodeentry=(inode_hello *)malloc(sizeof(inode_hello));
  
  inodeno=&ino;
  
  if(lfs_getattr(path,inodeattr)!=0)
     res = -ENOENT;
  
  if(inodeattr->type!=D)
    res = -ENOENT;
  
  char buffer[(inodeattr->size*sizeof(direntry))];
  
  entr=lfs_dirread(path,buffer,inodeattr->size);
  if(entr==-1){
    res = -ENOENT;
  }
  
  	
  icount=0;
  while (icount<entr) {
    struct stat st;
    memset(&st, 0, sizeof(st));
    memcpy((void *)name,(void *)(buffer+(icount*sizeof(direntry))),sizeof(char)*16);
    memcpy((void *)inodeno,(void *)(buffer+16+(icount*sizeof(direntry))),sizeof(char)*4);
    ino=(int)(*inodeno);
    if(read_inode(ino,inodentry,Rd)!=0){
      free(inodeattr);
      free(inodeentry);
      res = -ENOENT;
    }
    st.st_ino = ino;
    if(inodeentry->type==D)
      st.st_mode = 0755;
    else
      st.st_mode = 0666;
    
    if (filler(buf, name, &st, 0))
      break;
  }
  
  free(inodeattr);
  free(inodeentry);
  return res;
}
/***
fl_mkdir : Fuse interface to make a directory.

returns -1 on failure
         0 on success

***/

static int fl_mkdir(const char *path, mode_t mode)
{
	int res;

	
	if(lfs_filecreate(path,0755,D)!=0)
	  res = -ENOENT;

	return res;
}

/***

fl_unlink : Removing a file

return -1 on failure
        0 on success

***/
static int fl_unlink(const char *path)
{
  int res=0;
  if(lfs_fileunlink(path)!=0)
    res = -ENOENT;

  return res;
}

/*** 

fl_rmdir : Removing a Empty Directory


return -1 on failure
        0 on succcess

***/

static int fl_rmdir(const char *path)
{
  int res=0;
  
  if(fl_dirunlink(path)!=0)
    res = -ENOENT;
  return res;
}

/***
 fl_rename: File rename 


return -ENOENT on failure
        0 on Success

***/
static int fl_rename(const char *from, const char *to)
{
  int res=0,inodeno=2,parentfrom,parentto,block=-1,icount=0,incr=0,flag=-1;
  inode_hello *inoderm,*inodermp;
  char path[256],buf[SECTOR_SIZE]={0};
  inoderm=(inode_hello *)malloc(sizeof(inode_hello));
  inodermp=(inode_hello *)malloc(sizeof(inode_hello));
  char **temp=(char **)malloc(sizeof(char *)*300);
 
  inodeno=traverse_path(from);
  if(inodeno==-1)
    return -ENOENT;
         
  parentfrom=parent;
  
  strcpy((char *)path,(char *)to);
  if((strlen(path)==0)||(strlen(path)>255))
    return -ENOENT;
  temp[steps]=strtok(path1,"/");
  while(temp[steps]!=NULL){
    steps++;
    temp[steps]=strtok(NULL,"/");
  } 
  
  if(traverse_path(to)==-1 && parent==parentfrom){
    parentto=parent;
    if(read_inode(parentto,inodermp,Rd)!=0)
      return -ENOENT;
    
    while(icount<inodermp->size || flag!=1){
      block=inoderm->blockptr[blkcount];
      if(block==-1)
        break;
      if(Disk_Read(block,buf)!=0)
        return -1;
      for(intr=0;intr<26;intr++){
        if(strncmp(inoderm->name,buf+(sizeof(direntry)*intr),(sizeof(char)*16))==0){
          strcpy((char *)buf+(sizeof(direntry)*intr),(char *)temp[steps-1]);
          flag =1;
          break;
        }
	
      }
      
      icount++;
    }
    if(flag==1){
      if(Disk_Write(block,buf)!=0)
	return -ENOENT;
    }
    
    if(read_inode(inodeno,inoderm,Rd)!=0)
      return -ENOENT;
    
    strcpy((char *)inoderm->name,(char *)temp[steps-1]);
    
    if(read_inode(inodeno,inoderm,Up)!=0)
      return -ENOENT;
    

  }else{
    return -ENOENT;
  }
  free(inoderm);
  free(inodermp);
  return 0;
}
/***

fl_truncate : Truncates File 

if size>file size : appends with 0

if size<file size :truncate to new size

returns : -1 on failure
           0 on success

***/

static int fl_truncate(const char *path, off_t size)
{
  int res=0,trcflag=-1,blocks=0,bytes=0,icount=0,block=-1;
  inode_hello *inodetr=(inode_hello *)malloc(sizeof(inode_hello));
  int inodeno;
  char buf[SECTOR_SIZE]={0};

  if(lfs_getattr(path,inodetr)!=0)
    return -ENOENT;
  
  if(inodetr->size>size){
    trcflag=1;
  }else{
    trcflag=0;
  }

  inodeno=traverse_path(path);
  if(inodeno==-1)
    return -ENOENT;

  int fd=lfs_openfile(path);
  if(fd==-1)
    return -ENOENT;

  blocks = size/SECTOR_SIZE;
  bytes = size%SECTOR_SIZE;

  if(bytes==0)
    icount=blocks+1;
  else
    icount=blocks+2;
  
  if(trcflag==1){
    inodetr->size=size;
    if(icount<30){
      block=inodetr->blockptr[icount];
      if(bitmap_update(B,block,0,Up)!=0)
	return -ENOENT;
      if(Disk_Write(block,buf)!=0)
	return -ENOENT;
      inodetr->blockptr[icount]=-1;
      icount++;
    }
    if(read_inode(inodeno,inodetr,Up)!=0)
      return -ENOENT;
  }else{
    if(lfs_fileseek(fd,inodetr->size)!=0)
      return -ENOENT;
    blocks=(size-inodetr->size)/SECTOR_SIZE;
    bytes=(size-inodetr->size)%SECTOR_SIZE;
    icount=0;
    while(icount<blocks){
      if(lfs_writefile(fd,(void *)buf,SECTOR_SIZE)==-1){
	return -ENOENT;
      }
      icount++;
    }
    if(lfs_writefile(fd,(void *)buf,SECTOR_SIZE)==-1){
      return -ENOENT;
    }
    if(read_inode(inodeno,inodetr,Rd)!=0)
      return -ENOENT;
    
    inodetr->size=size;
    if(read_inode(inode,inodetr,Up)!=0)
      return -ENOENT;
  }
      
  free(inodetr);
  return 0;
}

/***

fl_open : Opening a file!

returns -1 failure
         fd on success

***/

static int fl_open(const char *path, struct fuse_file_info *fi)
{
	int res=0;

	res =lfs_openfile(path);
	if (res == -1)
	  return -ENOENT;

	if(lfs_close(res)==-1)
	  return -ENOENT;

       	return 0;
}

/*** 

fl_read: Read "size" number of bytes specified in the file


return -1 on failure 
       number of bytes read on success

***/

static int fl_read(const char *path, char *buf, size_t size, off_t offset,
		    struct fuse_file_info *fi)
{
  int fd,blocks=0,bytes=0;
  int res,icount=0;
  inode_hello *inoderd=(inode_hello *)malloc(sizeof(inode_hello));
    
  (void) fi;
  fd = lfs_openfile(path);
  if (fd == -1)
    return -ENOENT;
  
  if(lfs_fileseek(fd,offset)!=0)
    return -ENOENT;
  
  if(lfs_getattr(path,inoderd)!=0)
    return -ENOENT;

  if((offset+size)>(inoderd->size))
    size=offset+size;

  blocks=size/SECTOR_SIZE;
  bytes=size%SECTOR_SIZE;
  
  while(icount<blocks){
    res = lfs_readfile(fd, buf+(incr*SECTOR_SIZE), SECTOR_SIZE);
    if (res == -1)
      return -ENOENT;
    icount++;

  }
  memset(buf,0,SECTOR_SIZE);
  res = lfs_readfile(fd,buf+(SECTOR_SIZE*blocks),bytes);

  free(inoderd);
  if(lfs_close(fd)!=0)
    return -ENOENT;
  return size;
}

/***
    
fl_write : Writing size number of bytes to file 


returns -1 failure
         number of bytes written
	 
***/

static int fl_write(const char *path, const char *buf, size_t size,
		     off_t offset, struct fuse_file_info *fi)
{
  int fd,icount=0,blocks=0,bytes=0;
  int res;
  inode_hello *inodewr=(inode_hello *)malloc(sizeof(inode_hello));
  
  (void) fi;
  fd = lfs_openfile(path);
  if (fd == -1)
    return -ENOENT;
  
  if(lfs_fileseek(fd,offset)!=0)
    return -ENOENT;

  if(offset+size>MAXFILESIZE)
    return -ENOENT;

  blocks=size/SECTOR_SIZE;
  bytes=size%SECTOR_SIZE;
  while(icount<blocks){
    res = lfs_writefile(fd, buf+(icount*SECTOR_SIZE),SECTOR_SIZE);
    if (res == -1)
      return -ENOENT;
    icount++;
  }
  res=lfs_writefile(fd,buf+(blocks*SECTOR_SIZE),bytes);
  if(res==-1)
    return -ENOENT;
  if(lfs_close(fd)!=0)
    return -ENOENT;
  return size;
}

/***

fl_fsync : Flush on every close come what may!

return -1 on failure
        0 on success

***/
static int fl_fsync(const char *path, int isdatasync,
                     struct fuse_file_info *fi)
{
  if(FS_Sync("root/Desktop/fuse-2.7.3/example/newimple/test/DiskImage")!=0)
     return -ENOENT;
  return 0;
  
}



